@@ -158,7 +158,7 @@ day_of_the_week (stm *tm)
 	    + (corr_year / 4)
 	    - ((corr_year / 4) / 25) + ((corr_year / 4) % 25 < 0)
 	    + (((corr_year / 4) / 25) / 4)
-	    + __mon_yday[0][tm->tm_mon]
+	    + __mon_yday[0][tm->tm_mon] // no check on month in range
 	    + tm->tm_mday - 1);
     tm->tm_wday = ((wday % 7) + 7) % 7;
 }
@@ -180,6 +180,8 @@ day_of_the_year (stm *tm)
 #include <wchar.h>
 #include <wctype.h>
 
+#include <rlocale.h> // to possibly override iswspace
+
 #define DT_WBUFSIZE 25
 static wchar_t w_weekday_name[][DT_WBUFSIZE] =
 {
@@ -365,7 +367,7 @@ w_strptime_internal (wchar_t *rp, const wchar_t *fmt, stm *tm,
 	  break;
 	case L'j':
 	  /* Match day number of year.  */
-	  get_number (1, 366, 3);
+	  get_number (1, 366, 3); // NB: 366 would be invalid in most years
 	  tm->tm_yday = val - 1;
 	  have_yday = 1;
 	  break;
@@ -611,8 +613,8 @@ w_strptime_internal (wchar_t *rp, const wchar_t *fmt, stm *tm,
 	    case L'y':
 		/* Match year within century using alternate numeric symbols.  */
 		get_alt_number (0, 99, 2);
-	        int ival = val;
-	        tm->tm_year = ival >= 69 ? ival : ival + 100;
+		int ival = val;
+		tm->tm_year = ival >= 69 ? ival : ival + 100;
 		want_xday = 1;
 		break;
 	    default:
@@ -638,14 +640,22 @@ w_strptime_internal (wchar_t *rp, const wchar_t *fmt, stm *tm,
 
     if (want_xday && !have_wday) {
 	if ( !(have_mon && have_mday) && have_yday)  {
+	    /* have_yday, so this must have come from %j */
 	    /* We don't have tm_mon and/or tm_mday, compute them. */
 	    int t_mon = 0;
-	    while (__mon_yday[__isleap(1900 + tm->tm_year)][t_mon] <= tm->tm_yday)
-		t_mon++;
+	    int yr = 1900 + tm->tm_year;
+	    if(tm->tm_yday > (__isleap(yr) ? 365 : 364)) {
+		warning("day-of-year %d in year %d is invalid\n",
+			tm->tm_yday+1, yr);
+		t_mon = 12; // this will give an invalid mday, so invalid tm
+	    } else {
+		while (__mon_yday[__isleap(yr)][t_mon] <= tm->tm_yday)
+		    t_mon++;
+	    }
 	    if (!have_mon)
 		tm->tm_mon = t_mon - 1;
 	    if (!have_mday)
-		tm->tm_mday = (tm->tm_yday - __mon_yday[__isleap(1900 + tm->tm_year)][t_mon - 1] + 1);
+		tm->tm_mday = (tm->tm_yday - __mon_yday[__isleap(yr)][t_mon - 1] + 1);
 	}
 	day_of_the_week (tm);
     }
@@ -668,6 +678,8 @@ w_strptime_internal (wchar_t *rp, const wchar_t *fmt, stm *tm,
 	  tm->tm_mon = save_mon;
 
       if (!have_yday) {
+	  // Get yday from week and day-of-the-week.
+	  // This does not validate yday against any upper limit
 	  tm->tm_yday = ((7 - (tm->tm_wday - w_offset)) % 7
 			 + (week_no - 1) *7
 			 + save_wday - w_offset);
@@ -677,15 +689,20 @@ w_strptime_internal (wchar_t *rp, const wchar_t *fmt, stm *tm,
       if (!have_mday || !have_mon)
       {
 	  int t_mon = 0;
-	  while (__mon_yday[__isleap(1900 + tm->tm_year)][t_mon]
-		 <= tm->tm_yday)
-	      t_mon++;
+	  int yr = 1900 + tm->tm_year;
+	  if(tm->tm_yday > (__isleap(yr) ? 365 : 364)) {
+	      warning("(0-based) yday %d in year %d is invalid\n",
+		      tm->tm_yday, yr);
+	      t_mon = 12;
+	  } else {
+	      while (__mon_yday[__isleap(yr)][t_mon] <= tm->tm_yday)
+		  t_mon++;
+	  }
 	  if (!have_mon)
 	      tm->tm_mon = t_mon - 1;
 	  if (!have_mday)
 	      tm->tm_mday =
-		  (tm->tm_yday
-		   - __mon_yday[__isleap(1900 + tm->tm_year)][t_mon - 1] + 1);
+		  (tm->tm_yday - __mon_yday[__isleap(yr)][t_mon - 1] + 1);
       }
 
       tm->tm_wday = save_wday;
@@ -739,8 +756,8 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 
 	/* We need this for handling the `E' modifier.  */
     start_over:
-								/* #nocov start */ 
-	switch (*fmt++)                                      
+								/* #nocov start */
+	switch (*fmt++)
 	{
 	case '%':
 	    /* Match the `%' character itself.  */
@@ -792,7 +809,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	  get_number (0, 99, 2);
 	  century = val;
 	  want_xday = 1;
-	  break;
+	  break;                                                                /* #nocov end */
 	case 'd':
 	case 'e':
 	  /* Match day of month.  */
@@ -801,7 +818,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	  have_mday = 1;
 	  want_xday = 1;
 	  break;
-	case 'F':
+	case 'F':                                                                /* #nocov start */
 	  if (!recursive ("%Y-%m-%d"))
 	    return NULL;
 	  want_xday = 1;
@@ -814,14 +831,14 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	    return NULL;
 	  want_xday = 1;
 	  break;
-	case 'k':
+	case 'k':                                                                /* #nocov end */
 	case 'H':
 	  /* Match hour in 24-hour clock.  */
 	  get_number (0, 24, 2);  /* allow 24:00:00 */
 	  tm->tm_hour = val;
 	  have_I = 0;
 	  break;
-	case 'l':
+	case 'l':                                                                /* #nocov start */
 	  /* Match hour in 12-hour clock.  GNU extension.  */
 	case 'I':
 	  /* Match hour in 12-hour clock.  */
@@ -834,7 +851,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	  get_number (1, 366, 3);
 	  tm->tm_yday = val - 1;
 	  have_yday = 1;
-	  break;
+	  break;                                                                /* #nocov end */
 	case 'm':
 	  /* Match number of month.  */
 	  get_number (1, 12, 2);
@@ -847,7 +864,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	  get_number (0, 59, 2);
 	  tm->tm_min = val;
 	  break;
-	case 'n':
+	case 'n':                                                                /* #nocov start */
 	case 't':
 	  /* Match any white space.  */
 	  while (isspace ((int)*rp))
@@ -958,7 +975,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	    /* Indicate that we want to use the century, if specified.  */
 	    want_century = 1;
 	    want_xday = 1;
-	    break;
+	    break;                                                                /* #nocov end */
 	case 'Y':
 	    /* Match year including century number.  */
 	    get_number (0, 9999, 4);
@@ -966,7 +983,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 	    want_century = 0;
 	    want_xday = 1;
 	    break;
-	case 'z':
+	case 'z':                                                                /* #nocov start */
 	    /* Only recognize RFC 822 form */
 	    {
 		int n = 0, neg, off = 0;
@@ -1039,7 +1056,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 		/* Match minutes using alternate numeric symbols.  */
 		get_alt_number (0, 59, 2);
 		tm->tm_min = val;
-		break;
+		break;                                                                /* #nocov end */
 	    case 'S':
 		/* Match seconds using alternate numeric symbols.
 		   get_alt_number (0, 61, 2); */
@@ -1054,7 +1071,7 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 		       rp = end;
 		   }
 		break;
-	    case 'U':
+	    case 'U':                                                                /* #nocov start */
 	      get_alt_number (0, 53, 2);
 	      week_no = val;
 	      have_uweek = 1;
@@ -1105,14 +1122,22 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
 
     if (want_xday && !have_wday) {
 	if ( !(have_mon && have_mday) && have_yday)  {
+	    /* have_yday, so this must have come from %j */
 	    /* We don't have tm_mon and/or tm_mday, compute them. */
 	    int t_mon = 0;
-	    while (__mon_yday[__isleap(1900 + tm->tm_year)][t_mon] <= tm->tm_yday)
-		t_mon++;
+	    int yr = 1900 + tm->tm_year;
+	    if(tm->tm_yday > (__isleap(yr) ? 365 : 364)) {
+		warning("day-of-year %d in year %d is invalid\n",
+			tm->tm_yday+1, yr);
+		t_mon = 12;
+	    } else {
+		while (__mon_yday[__isleap(yr)][t_mon] <= tm->tm_yday)
+		    t_mon++;
+	    }
 	    if (!have_mon)
 		tm->tm_mon = t_mon - 1;
 	    if (!have_mday)
-		tm->tm_mday = (tm->tm_yday - __mon_yday[__isleap(1900 + tm->tm_year)][t_mon - 1] + 1);
+		tm->tm_mday = (tm->tm_yday - __mon_yday[__isleap(yr)][t_mon - 1] + 1);
 	}
 	day_of_the_week (tm);
     }
@@ -1142,20 +1167,25 @@ strptime_internal (const char *rp, const char *fmt, stm *tm,
       if (!have_mday || !have_mon)
       {
 	  int t_mon = 0;
-	  while (__mon_yday[__isleap(1900 + tm->tm_year)][t_mon]
-		 <= tm->tm_yday)
-	      t_mon++;
+	  int yr = 1900 + tm->tm_year;
+	  if(tm->tm_yday > (__isleap(yr) ? 365 : 364)) {
+		warning("(0-based) yday %d in year %d is invalid\n",
+			tm->tm_yday, yr);
+		t_mon = 12;
+	  } else {
+	      while (__mon_yday[__isleap(yr)][t_mon] <= tm->tm_yday)
+		  t_mon++;
+	  }
 	  if (!have_mon)
 	      tm->tm_mon = t_mon - 1;
 	  if (!have_mday)
 	      tm->tm_mday =
-		  (tm->tm_yday
-		   - __mon_yday[__isleap(1900 + tm->tm_year)][t_mon - 1] + 1);
+		  (tm->tm_yday - __mon_yday[__isleap(yr)][t_mon - 1] + 1);
       }
 
-      tm->tm_wday = save_wday;
+      tm->tm_wday = save_wday;                                                                /* #nocov end */
   }
-								/* #nocov end */ 
+
     return (char *) rp;
 }
 
@@ -1256,7 +1286,7 @@ static void get_locale_w_strings(void)
 
 /* We only care if the result is null or not */
 static void *
-R_strptime (const char *buf, const char *format, stm *tm, 
+R_strptime (const char *buf, const char *format, stm *tm,
 	    double *psecs, int *poffset)
 {
 #if defined(HAVE_WCSTOD)
