@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2000-2017  The R Core Team.
+ *  Copyright (C) 2000-2021  The R Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -37,7 +37,7 @@
 # include <config.h>
 #endif
 
-#include <Rmath.h> // Rexp10
+#include <Rmath.h> // for imin2()
 
 // to get tm_zone, tm_gmtoff defined in glibc.
 // some other header, e.g. math.h, might define the macro.
@@ -84,7 +84,6 @@ known OS with 64-bit time_t and complete tables is Linux.
 */
 
 
-char *R_tzname[2];
 // #define USE_INTERNAL_MKTIME
 #define HAVE_POSIX_LEAPSECONDS
 
@@ -102,6 +101,14 @@ char *R_tzname[2];
 # define HAVE_WORKING_64BIT_MKTIME 1
 #else
 
+# if defined(__CYGWIN__) || defined(_WIN32)
+  // nothing, in particular so not define GMTOFF
+# else
+  // define this to have the field, even if empty / NA
+  # undef HAVE_TM_GMTOFF
+  # define HAVE_TM_GMTOFF 1
+#endif
+
 typedef struct tm stm;
 #define R_tzname tzname
 # if defined(__CYGWIN__) || defined(_WIN32)
@@ -116,7 +123,7 @@ extern char *tzname[2];
 /*#include <Defn.h>*/
 /*#include <Internal.h>*/
 #include <string.h>
-#define attribute_hidden  
+#define attribute_hidden
 #define HAVE_PUTENV
 #define Rexp10(x) pow(10.0, x) /* DEdd */
 #include <Rinternals.h>
@@ -144,7 +151,7 @@ static int validate_tm (stm *tm)
 	res++; 							/* #nocov start */
 	tmp = tm->tm_sec/60;
 	tm->tm_sec -= 60 * tmp; tm->tm_min += tmp;
-	if(tm->tm_sec < 0) {tm->tm_sec += 60; tm->tm_min--;}
+	if(tm->tm_sec < 0) {tm->tm_sec += 60; tm->tm_min--;}	/* #nocov end */
     }
 
     if (tm->tm_min < 0 || tm->tm_min > 59) {
@@ -246,10 +253,10 @@ static double mktime00 (stm *tm)
     if (year0 > 1970) {
 	for (year = 1970; year < year0; year++)
 	    day += days_in_year(year);
-    } else if (year0 < 1970) {
+    } else if (year0 < 1970) {                                   /* #nocov start */
 	for (year = 1969; year >= year0; year--)
 	    day -= days_in_year(year);
-    }
+    }								/* #nocov end */
 
     /* weekday: Epoch day was a Thursday */
     if ((tm->tm_wday = (day + 4) % 7) < 0) tm->tm_wday += 7;
@@ -322,7 +329,7 @@ static const time_t leapseconds[] = // dput(unclass(.leap.seconds)) :
    915148800,1136073600,1230768000,1341100800,1435708800,1483228800};
 #endif
 
-static double guess_offset (stm *tm)					/* #nocov start */ 
+static double guess_offset (stm *tm)					/* #nocov start */
 {
     double offset, offset1, offset2;
     int i, wday, year, oldmonth, oldisdst, oldmday;
@@ -439,7 +446,7 @@ static double mktime0 (stm *tm, const int local)
 	return res;
 /* watch the side effect here: both calls alter their arg */
     } else return guess_offset(tm) + mktime00(tm);
-}								/* #nocov end */ 
+}								/* #nocov end */
 
 /* Interface for localtime or gmtime or internal substitute */
 static stm * localtime0(const double *tp, const int local, stm *ltm)
@@ -475,7 +482,7 @@ static stm * localtime0(const double *tp, const int local, stm *ltm)
 
     /* internal substitute code.
        Like localtime, this returns a pointer to a static struct tm */
-									/* #nocov start */ 
+									/* #nocov start */
     int day = (int) floor(d/86400.0);
     int left = (int) (d - day * 86400.0 + 1e-6); // allow for fractional secs
 
@@ -547,7 +554,7 @@ static stm * localtime0(const double *tp, const int local, stm *ltm)
     } else {
 	res->tm_isdst = 0; /* no dst in GMT */
 	return res;
-    }								/* #nocov end */ 
+    }								/* #nocov end */
 }
 #endif
 
@@ -606,7 +613,7 @@ static void reset_tz(char *tz)
     tzset();
 }
 
-static void glibc_fix(stm *tm, int *invalid) 			/* #nocov start */ 
+static void glibc_fix(stm *tm, int *invalid) 			/* #nocov start */
 {
     /* set mon and mday which glibc does not always set.
        Use current year/... if none has been specified.
@@ -647,7 +654,7 @@ static void glibc_fix(stm *tm, int *invalid) 			/* #nocov start */
 	}
 	if(tm->tm_mon == NA_INTEGER) tm->tm_mon = tm0->tm_mon;
     }
-}								/* #nocov end */ 
+}								/* #nocov end */
 
 
 static const char ltnames [][7] =
@@ -681,6 +688,7 @@ makelt(stm *tm, SEXP ans, R_xlen_t i, int valid, double frac_secs)
 
 // We assume time zone names/abbreviations are ASCII, as all known ones are.
 
+// .Internal(as.POSIXlt(x, tz)) -- called only from  as.POSIXlt.POSIXct()
 //SEXP attribute_hidden do_asPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
 {
@@ -749,7 +757,7 @@ SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
 	stm dummy, *ptm = &dummy;
 	double d = REAL(x)[i];
 	if(R_FINITE(d)) {
-	    ptm = localtime0(&d, 1 - isgmt, &dummy);
+	    ptm = localtime0(&d, !isgmt, &dummy);
 	    /* in theory localtime/gmtime always return a valid
 	       struct tm pointer, but Windows uses NULL for error
 	       conditions (like negative times). */
@@ -758,7 +766,7 @@ SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
 	makelt(ptm, ans, i, valid, d - floor(d));
 	if(!isgmt) {
 	    char *p = "";
-	    // or ptm->tm_zone
+	    // or ptm->tm_zone (but not specified by POSIX)
 	    if(valid && ptm->tm_isdst >= 0)
 		p = R_tzname[ptm->tm_isdst];
 	    SET_STRING_ELT(VECTOR_ELT(ans, 9), i, mkChar(p));
@@ -777,7 +785,7 @@ SEXP asPOSIXlt(SEXP argsxp, SEXP tzarg) // other args?
     SEXP nm = getAttrib(x, R_NamesSymbol);
     if(nm != R_NilValue) setAttrib(VECTOR_ELT(ans, 5), R_NamesSymbol, nm);
     if(settz) reset_tz(oldtz);
-    UNPROTECT(5);
+    unprotect(5);               /* DEdd: One commented out so 5 instead of 6 */
     return ans;
 }
 
@@ -794,7 +802,7 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
 
     /* checkArity(op, args); */
     /* PROTECT(x = duplicate(CAR(args))); /\* coerced below *\/ */
-    PROTECT(x = duplicate(sxparg)); /* coerced below */ 
+    PROTECT(x = duplicate(sxparg)); /* coerced below */
     if (!isVectorList(x) || LENGTH(x) < 9)
         error("invalid '%s' argument", "x");			/* #nocov */
     /* if(!isString((stz = CADR(args))) || LENGTH(stz) != 1) */
@@ -855,14 +863,14 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
 	    REAL(ans)[i] = NA_REAL;					/* #nocov */
 	else {
 	    errno = 0;
-	    tmp = mktime0(&tm, 1 - isgmt);
+	    tmp = mktime0(&tm, !isgmt);
 #ifdef MKTIME_SETS_ERRNO
 	    REAL(ans)[i] = errno ? NA_REAL : tmp + (secs - fsecs);
 #else
 	    REAL(ans)[i] = ((tmp == -1.)
 			    /* avoid silly gotcha at epoch minus one sec */
 			    && (tm.tm_sec != 59)					/* #nocov */
-			    && ((tm.tm_sec = 58), (mktime0(&tm, 1 - isgmt) != -2.))	/* #nocov */
+			    && ((tm.tm_sec = 58), (mktime0(&tm, !isgmt) != -2.))	/* #nocov */
 			    ) ?
 	      NA_REAL : tmp + (secs - fsecs);
 #endif
@@ -876,7 +884,7 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
     SET_STRING_ELT(klass, 0, mkChar("POSIXct"));
     SET_STRING_ELT(klass, 1, mkChar("POSIXt"));
     classgets(ans, klass);
-    
+
     UNPROTECT(4);
     return ans;
 }
@@ -884,8 +892,8 @@ SEXP asPOSIXct(SEXP sxparg, SEXP tzarg) //
 //SEXP attribute_hidden do_formatPOSIXlt(SEXP call, SEXP op, SEXP args, SEXP env)
 SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 {
-    SEXP x, sformat, ans, tz;
-    R_xlen_t n = 0, m, N, nlen[9];
+    SEXP x, sformat, tz;
+    R_xlen_t m;
     int UseTZ, settz = 0;
     char buff[300];
     char oldtz[1001] = "";
@@ -893,7 +901,7 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 
     /* checkArity(op, args); */
     /* PROTECT(x = duplicate(CAR(args))); /\* coerced below *\/ */
-    PROTECT(x = duplicate(argsxp)); /* coerced below */ 
+    PROTECT(x = duplicate(argsxp)); /* coerced below */
     if (!isVectorList(x) || LENGTH(x) < 9)
         error("invalid '%s' argument", "x");					/* #nocov */
     /* if(!isString((sformat = CADR(args))) || XLENGTH(sformat) == 0) */
@@ -915,6 +923,9 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 	    const char *p = translateChar(STRING_ELT(sformat, i));
 	    if (strstr(p, "%Z") || strstr(p, "%z")) {needTZ = 1; break;}
 	}
+	/* strftime (per POSIX) calls settz(), so we need to set TZ, but
+	   we would not have to call settz() directly (except for the
+	   old OLD_Win32 code) */
 	if(needTZ) settz = set_tz(tz1, oldtz);
     }
 
@@ -924,29 +935,33 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
     memset(&tm, 0, sizeof(tm));
 
     /* coerce fields, find length of longest one */
-    for(int i = 0; i < 9; i++) {
+    R_xlen_t n = 0, nlen[11];
+    int nn = imin2(LENGTH(x), 11);
+    for(int i = 0; i < nn; i++) {
 	nlen[i] = XLENGTH(VECTOR_ELT(x, i));
 	if(nlen[i] > n) n = nlen[i];
-	// real for 'sec', the first; integer for the rest:
-	SET_VECTOR_ELT(x, i, coerceVector(VECTOR_ELT(x, i),
-					  i > 0 ? INTSXP : REALSXP));
+	if(i != 9) // real for 'sec', the first; integer for the rest:
+	    SET_VECTOR_ELT(x, i, coerceVector(VECTOR_ELT(x, i),
+					      i > 0 ? INTSXP : REALSXP));
     }
     if(n > 0) {
-	for(int i = 0; i < 9; i++)
+	for(int i = 0; i < nn; i++)
 	    if(nlen[i] == 0)
 		error("zero-length component in non-empty \"POSIXlt\" structure");	/* #nocov */
     }
-    if(n > 0) N = (m > n) ? m:n; else N = 0;
-    PROTECT(ans = allocVector(STRSXP, N));
+    R_xlen_t N = (n > 0) ? ((m > n) ? m : n) : 0;
+    SEXP ans = PROTECT(allocVector(STRSXP, N));
     char tm_zone[20];
 #ifdef HAVE_TM_GMTOFF
-    Rboolean have_zone = LENGTH(x) >= 11 && XLENGTH(VECTOR_ELT(x, 9)) == n &&
-	XLENGTH(VECTOR_ELT(x, 10)) == n;
+    Rboolean have_zone = LENGTH(x) >= 11;// and components w/ length >= 1
 #else
-    Rboolean have_zone = LENGTH(x) >= 10 && XLENGTH(VECTOR_ELT(x, 9)) == n;
+    Rboolean have_zone = LENGTH(x) >= 10;
 #endif
     if(have_zone && !isString(VECTOR_ELT(x, 9)))
 	error("invalid component [[10]] in \"POSIXlt\" should be 'zone'");		/* #nocov */
+    if(!have_zone && LENGTH(x) > 9) // rather even error ?
+        /* never when !HAVE_GMTOFF */
+	warning("More than 9 list components in \"POSIXlt\" without timezone");		/* #nocov */
     for(R_xlen_t i = 0; i < N; i++) {
 	double secs = REAL(VECTOR_ELT(x, 0))[i%nlen[0]], fsecs = floor(secs);
 	// avoid (int) NAN
@@ -959,19 +974,22 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 	tm.tm_wday  = INTEGER(VECTOR_ELT(x, 6))[i%nlen[6]];
 	tm.tm_yday  = INTEGER(VECTOR_ELT(x, 7))[i%nlen[7]];
 	tm.tm_isdst = INTEGER(VECTOR_ELT(x, 8))[i%nlen[8]];
-	if(have_zone) {
-	    strncpy(tm_zone, CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%n)), 20);
+	if(have_zone) { // not "UTC", e.g.
+	    strncpy(tm_zone, CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%nlen[9])), 20 - 1);
 	    tm_zone[20 - 1] = '\0';
 #ifdef HAVE_TM_ZONE
 	    tm.tm_zone = tm_zone;
 #elif defined USE_INTERNAL_MKTIME
 	    if(tm.tm_isdst >= 0) R_tzname[tm.tm_isdst] = tm_zone;
 #else
-	    // The system one, as we use system strftime here
-	    if(tm.tm_isdst >= 0) tzname[tm.tm_isdst] = tm_zone;
+	    /* Modifying tzname causes memory corruption on Solaris. It
+	       is not specified to have any effect and strftime is documented
+	       to call settz().*/
+	    if(tm.tm_isdst >= 0 && strcmp(tzname[tm.tm_isdst], tm_zone))
+		warning("Timezone specified in the object field cannot be used on this system."); /* #nocov */
 #endif
 #ifdef HAVE_TM_GMTOFF
-	    int tmp = INTEGER(VECTOR_ELT(x, 10))[i%n];
+	    int tmp = INTEGER(VECTOR_ELT(x, 10))[i%nlen[10]];
 	    tm.tm_gmtoff = (tmp == NA_INTEGER) ? 0 : tmp;
 #endif
 	}
@@ -1033,13 +1051,14 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 	    strftime(buff, 256, buf2, &tm);
 #endif
 	    buff[256] = '\0';
+            //mbcsTruncateToValid(buff);
 	    // Now assume tzone abbreviated name is < 40 bytes,
 	    // but they are currently 3 or 4 bytes.
 	    if(UseTZ) {
 		if(have_zone) {
-		    const char *p = CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%n));
+		    const char *p = CHAR(STRING_ELT(VECTOR_ELT(x, 9), i%nlen[9]));
 		    if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
-		} else if(!isNull(tz)) {		/* #nocov start */
+		} else if(!isNull(tz)) {                                        /* #nocov start */
 		    int ii = 0;
 		    if(LENGTH(tz) == 3) {
 			if(tm.tm_isdst > 0) ii = 2;
@@ -1048,7 +1067,7 @@ SEXP formatPOSIXlt(SEXP argsxp, SEXP fmtsxp, SEXP tzsxp) //
 		    }
 		    const char *p = CHAR(STRING_ELT(tz, ii));
 		    if(strlen(p)) {strcat(buff, " "); strcat(buff, p);}
-		}					/* #nocov end */
+		}								/* #nocov end */
 	    }
 	    SET_STRING_ELT(ans, i, mkChar(buff));
 	}
@@ -1081,14 +1100,14 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
         error("invalid '%s' argument", "x"); 				/* #nocov */
     /* if(!isString((sformat = CADR(args))) || XLENGTH(sformat) == 0) */
     /*     error(_("invalid '%s' argument"), "x"); */
-    if (!isString(sformat) || XLENGTH(sformat) == 0) 
+    if (!isString(sformat) || XLENGTH(sformat) == 0)
         error("invalid '%s' argument", "sformat");  			/* #nocov */
     /* if(!isString((stz = CADDR(args))) || LENGTH(stz) != 1) */
     /*     error(_("invalid '%s' value"), "tz"); */
-    if (!isString(stz) || LENGTH(stz) != 1) 
+    if (!isString(stz) || LENGTH(stz) != 1)
          error("invalid '%s' value", "tz");  				/* #nocov */
-    tz = CHAR(STRING_ELT(stz, 0)); 
-    if (strlen(tz) == 0) {
+    tz = CHAR(STRING_ELT(stz, 0));
+    if(strlen(tz) == 0) {
 	/* do a direct look up here as this does not otherwise
 	   work on Windows */
 	char *p = getenv("TZ");
@@ -1129,9 +1148,9 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
     for(int i = 0; i < 9; i++)
 	SET_VECTOR_ELT(ans, i, allocVector(i > 0 ? INTSXP : REALSXP, N));
     if(!isgmt) {
-	SET_VECTOR_ELT(ans, 9, allocVector(STRSXP, n));
+	SET_VECTOR_ELT(ans, 9, allocVector(STRSXP, N));
 #ifdef HAVE_TM_GMTOFF
-	SET_VECTOR_ELT(ans, 10, allocVector(INTSXP, n));
+	SET_VECTOR_ELT(ans, 10, allocVector(INTSXP, N));
 #endif
     }
 
@@ -1165,7 +1184,7 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
 	    tm.tm_isdst = -1;
 	    if (offset != NA_INTEGER) {
 #ifdef HAVE_TM_GMTOFF
-		tm.tm_gmtoff = offset;
+		tm.tm_gmtoff = offset;					/* #nocov */
 #endif
 		/* we know the offset, but not the timezone
 		   so all we can do is to convert to time_t,
@@ -1181,7 +1200,7 @@ SEXP Rstrptime(SEXP xarg, SEXP sformatarg, SEXP stzarg) {
 		/* we do want to set wday, yday, isdst, but not to
 		   adjust structure at DST boundaries */
 		memcpy(&tm2, &tm, sizeof(stm));
-		mktime0(&tm2, 1-isgmt); /* set wday, yday, isdst */
+		mktime0(&tm2, !isgmt); /* set wday, yday, isdst */
 		tm.tm_wday = tm2.tm_wday;
 		tm.tm_yday = tm2.tm_yday;
 		tm.tm_isdst = isgmt ? 0: tm2.tm_isdst;
@@ -1230,10 +1249,10 @@ SEXP D2POSIXlt(SEXP argsxp)
     /* checkArity(op, args); */
     /* PROTECT(x = coerceVector(CAR(args), REALSXP)); */
     PROTECT(x = coerceVector(argsxp, REALSXP));
-    n = XLENGTH(x); 
-    PROTECT(ans = allocVector(VECSXP, 9)); 
-    for(int i = 0; i < 9; i++) 
-        SET_VECTOR_ELT(ans, i, allocVector(i > 0 ? INTSXP : REALSXP, n)); 
+    n = XLENGTH(x);
+    PROTECT(ans = allocVector(VECSXP, 9));
+    for(int i = 0; i < 9; i++)
+	SET_VECTOR_ELT(ans, i, allocVector(i > 0 ? INTSXP : REALSXP, n));
 
     PROTECT(ansnames = allocVector(STRSXP, 9));
     for(int i = 0; i < 9; i++)
@@ -1292,8 +1311,8 @@ SEXP POSIXlt2D(SEXP sxparg)
 
     /* checkArity(op, args); */
     /* PROTECT(x = duplicate(CAR(args))); */
-    PROTECT(x = duplicate(sxparg)); 
-    if(!isVectorList(x) || LENGTH(x) < 9) 
+    PROTECT(x = duplicate(sxparg));
+    if(!isVectorList(x) || LENGTH(x) < 9)
         error("invalid '%s' argument", "x");  				/* #nocov */
 
     for(int i = 3; i < 6; i++)
